# Copyright 2023 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for pyglove.core.typing.class_schema."""

import copy
import inspect
import sys
from typing import Optional, Union, List
import unittest

from pyglove.core.typing import annotation_conversion  # pylint: disable=unused-import
from pyglove.core.typing import class_schema
from pyglove.core.typing import custom_typing
from pyglove.core.typing import key_specs as ks
from pyglove.core.typing import typed_missing
from pyglove.core.typing import value_specs as vs
from pyglove.core.typing.class_schema import Field
from pyglove.core.typing.class_schema import Schema


class ForwardRefTest(unittest.TestCase):
  """Test for `ForwardRef` class."""

  def setUp(self):
    super().setUp()
    self._module = sys.modules[__name__]

  def test_basics(self):
    r = class_schema.ForwardRef(self._module, 'FieldTest')
    self.assertIs(r.module, self._module)
    self.assertEqual(r.name, 'FieldTest')
    self.assertEqual(r.qualname, f'{self._module.__name__}.FieldTest')

  def test_resolved(self):
    self.assertTrue(class_schema.ForwardRef(self._module, 'FieldTest').resolved)
    self.assertFalse(class_schema.ForwardRef(self._module, 'Foo').resolved)

  def test_as_annotation(self):
    self.assertEqual(
        class_schema.ForwardRef(self._module, 'FieldTest').as_annotation(),
        FieldTest,
    )
    self.assertEqual(
        class_schema.ForwardRef(self._module, 'Foo').as_annotation(), 'Foo'
    )

  def test_cls(self):
    self.assertIs(
        class_schema.ForwardRef(self._module, 'FieldTest').cls, FieldTest
    )

    with self.assertRaisesRegex(TypeError, '.* does not exist in module'):
      _ = class_schema.ForwardRef(self._module, 'Foo').cls

    with self.assertRaisesRegex(TypeError, '.* is not a class'):
      _ = class_schema.ForwardRef(self._module, 'unittest').cls

  def test_format(self):
    self.assertEqual(
        str(class_schema.ForwardRef(self._module, 'FieldTest')),
        f'ForwardRef(module={self._module.__name__}, name=FieldTest)',
    )

  def test_eq_ne(self):
    ref = class_schema.ForwardRef(self._module, 'FieldTest')
    self.assertEqual(ref, ref)
    self.assertEqual(ref, class_schema.ForwardRef(self._module, 'FieldTest'))
    self.assertEqual(ref, FieldTest)
    self.assertEqual(FieldTest, ref)

    self.assertNotEqual(int, ref)
    self.assertNotEqual(ref, class_schema.ForwardRef(unittest, 'FieldTest'))
    self.assertNotEqual(ref, class_schema.ForwardRef(self._module, 'Foo'))


class FieldTest(unittest.TestCase):
  """Test for `Field` class."""

  def test_basics(self):
    f = Field('a', vs.Int(max_value=10), 'a field')
    self.assertIsInstance(f.key, ks.ConstStrKey)
    self.assertEqual(f.key, 'a')
    self.assertEqual(f.value, vs.Int(max_value=10))
    self.assertFalse(f.frozen)
    self.assertIs(f.annotation, int)
    self.assertEqual(f.default_value, typed_missing.MISSING_VALUE)
    self.assertEqual(f.description, 'a field')
    self.assertIsInstance(f.metadata, dict)
    self.assertEqual(len(f.metadata), 0)

    # Cover the self comparison in __eq__.
    self.assertEqual(f, f)

    # Test field with metadata.
    f = Field('a', vs.Bool(), 'a field', {'meta1': 'foo'})
    self.assertEqual(f.metadata['meta1'], 'foo')

    with self.assertRaisesRegex(ValueError, 'metadata must be a dict.'):
      Field('a', vs.Bool(), 'a field', 'abc')

  def test_extend(self):
    # Extend value spec and description.
    self.assertEqual(
        Field('a', vs.Int(1)).extend(
            Field('a', vs.Int(min_value=0), 'field a')),
        Field('a', vs.Int(min_value=0, default=1), 'field a'))

    # Overrided description will not inherit.
    self.assertEqual(
        Field('a', vs.Int(1), 'overrided field a').extend(
            Field('a', vs.Int(min_value=0), 'field a')),
        Field('a', vs.Int(min_value=0, default=1), 'overrided field a'))

    # Extend metadata.
    self.assertEqual(
        Field('a', vs.Int(), None, {
            'b': 1
        }).extend(Field('a', vs.Int(), 'field a', {'a': 2})),
        Field('a', vs.Int(), 'field a', {
            'a': 2,
            'b': 1
        }))

    # Extend with base field with different key is not allowed.
    with self.assertRaisesRegex(
        KeyError, '.* cannot extend .* for keys are different.'):
      Field('a', vs.Int()).extend(Field('b', vs.Int()))

  def test_apply(self):
    self.assertEqual(Field('a', vs.Int()).apply(1), 1)
    self.assertEqual(
        Field('a', vs.Dict([
            ('b', vs.Int()),
            ('c', vs.Bool())
        ])).apply({
            'b': 1,
            'c': True
        }), {
            'b': 1,
            'c': True
        })

    self.assertEqual(
        Field('a', vs.Dict([
            ('b', vs.Int()),
            ('c', vs.Bool())
        ])).apply({
            'b': 1,
        }, allow_partial=True), {
            'b': 1,
            'c': typed_missing.MISSING_VALUE
        })

    with self.assertRaisesRegex(
        TypeError, 'Expect <(type|class) \'int\'> but encountered '
        '<(type|class) \'float\'>'):
      Field('a', vs.Int()).apply(1.0)

    with self.assertRaisesRegex(
        ValueError, 'Required value is not specified. \\(Path=\'c\'.*\\)'):
      Field('a', vs.Dict([
          ('b', vs.Int()),
          ('c', vs.Bool())
      ])).apply({
          'b': 1,
      }, allow_partial=False)

  def test_format(self):
    self.assertEqual(
        Field('a', vs.Dict([
            ('b', vs.Int())
        ]), 'this is a very long field.', {
            'm1': 1,
            'm2': 2,
            'm3': 3,
            'm4': 4,
            'm5': 5
        }).format(compact=True, verbose=False),
        'Field(key=a, value=Dict({b=Int()}), '
        'description=\'this is a very long ...\', '
        'metadata={...})')

    self.assertEqual(
        Field('a', vs.Dict([
            ('b', vs.Int())
        ]), 'this is a very long field.', {
            'm1': 1,
            'm2': 2,
            'm3': 3,
            'm4': 4,
            'm5': 5
        }).format(compact=True, verbose=True),
        'Field(key=a, value=Dict({b=Int()}), '
        'description=\'this is a very long field.\', '
        'metadata={\'m1\': 1, \'m2\': 2, \'m3\': 3, \'m4\': 4, \'m5\': 5})')

    self.assertEqual(
        Field('a', vs.Dict([
            ('b', vs.Int())
        ]), 'field a').format(compact=False, verbose=False),
        'Field(key=a, value=Dict({\n'
        '    b = Int()\n'
        '  }), description=\'field a\')')


class SimpleObject:
  """Simple object for testing."""


class SchemaTest(unittest.TestCase):
  """Tests for Schema class."""

  def _create_test_schema(self, init_arg_list=None):
    return Schema([
        Field('a', vs.Int(1), 'Field a.'),
        Field('b', vs.Bool().noneable(), 'Field b.'),
        Field('c', vs.Dict([
            Field('d', vs.List(
                vs.Enum(0, [0, 1, None]), default=[0, 1]), 'Field d.'),
            Field('e', vs.List(
                vs.Dict([
                    (ks.StrKey(regex='foo.*'), vs.Str(),
                     'Mapped values.')
                ])
            ), 'Field e.'),
            Field('f', vs.Object(SimpleObject), 'Field f.')
        ]).noneable(), 'Field c.'),
    ], metadata={'init_arg_list': init_arg_list or []})

  def test_basics(self):
    s = Schema([Field('a', vs.Int())], 'schema1',
               [class_schema.create_schema([('b', vs.Bool())])])

    # Test Schema.fields.
    self.assertEqual(
        s.fields,
        {
            ks.ConstStrKey('b'): Field(ks.ConstStrKey('b'), vs.Bool()),
            ks.ConstStrKey('a'): Field(ks.ConstStrKey('a'), vs.Int()),
        })

    # Test Schema.__getitem__ and Schema.__contains__.
    self.assertEqual(s['a'], Field('a', vs.Int()))
    with self.assertRaises(KeyError):
      _ = s['foo']
    self.assertEqual(s.get('a'), Field('a', vs.Int()))
    self.assertIsNone(s.get('foo'))
    self.assertIn('a', s)
    self.assertNotIn('foo', s)

    # Test Schema.keys and Schema.values
    self.assertEqual(
        list(s.keys()), [ks.ConstStrKey('b'), ks.ConstStrKey('a')])
    self.assertEqual(
        list(s.values()),
        [Field(ks.ConstStrKey('b'), vs.Bool()),
         Field(ks.ConstStrKey('a'), vs.Int())])

    # Test Schema.allow_nonconst_keys
    self.assertFalse(s.allow_nonconst_keys)

    # Test Schema.name
    self.assertEqual(s.name, 'schema1')

    # Test Schema.set_name
    s.set_name('schema2')
    self.assertEqual(s.name, 'schema2')

    # Test Schema.description.
    self.assertIsNone(s.description)

    # Test Schema.set_description.
    s.set_description('schema1')
    self.assertEqual(s.description, 'schema1')

    # Test equal.
    self.assertEqual(s, s)
    self.assertNotEqual(s, class_schema.create_schema([]))

    with self.assertRaisesRegex(
        TypeError, 'Argument \'fields\' must be a list.'):
      Schema({'a': vs.Int()})

    with self.assertRaisesRegex(
        ValueError, 'NonConstKey is not allowed in schema'):
      Schema([Field(ks.StrKey(), vs.Int())], allow_nonconst_keys=False)

  def test_metadata(self):
    self.assertEqual(Schema([]).metadata, {})
    self.assertEqual(
        Schema([], metadata={
            'meta1': 1
        }).metadata, {'meta1': 1})

  def test_eq(self):
    self.assertEqual(self._create_test_schema(), self._create_test_schema())

  def test_format(self):
    """Tests for Schema.format."""
    self.assertEqual(
        self._create_test_schema().format(compact=True),
        'Schema(a=Int(default=1), b=Bool(default=None, noneable=True), '
        'c=Dict({d=List('
        'Enum(default=0, values=[0, 1, None]), default=[0, 1]), '
        'e=List('
        'Dict({StrKey(regex=\'foo.*\')=Str()})), '
        'f=Object(SimpleObject)}, noneable=True))')

    self.assertEqual(
        inspect.cleandoc(self._create_test_schema().format(
            compact=False, verbose=False)),
        inspect.cleandoc("""Schema(
          a = Int(default=1),
          b = Bool(default=None, noneable=True),
          c = Dict({
            d = List(Enum(default=0, values=[0, 1, None]), default=[0, 1]),
            e = List(Dict({
              StrKey(regex=\'foo.*\') = Str()
            })),
            f = Object(SimpleObject)
          }, noneable=True)
        )"""))

    self.assertEqual(
        inspect.cleandoc(
            # Equal to schema.format(compact=False, verbose=True)
            str(self._create_test_schema())),
        inspect.cleandoc("""Schema(
            # Field a.
            a = Int(default=1),

            # Field b.
            b = Bool(default=None, noneable=True),

            # Field c.
            c = Dict({
              # Field d.
              d = List(Enum(default=0, values=[0, 1, None]), default=[0, 1]),

              # Field e.
              e = List(Dict({
                # Mapped values.
                StrKey(regex=\'foo.*\') = Str()
              })),

              # Field f.
              f = Object(SimpleObject)
            }, noneable=True)
          )"""))

  def test_extend(self):
    """Tests for Schema.extend."""

    # Disjoint inheirtance with preserved field order.
    self.assertEqual(
        list(
            class_schema.create_schema([
                ('a', vs.Int()), ('c', vs.Dict([('d', vs.Str())]))
            ]).extend(
                class_schema.create_schema([
                    ('b', vs.Bool().noneable()),
                    ('c', vs.Dict([
                        ('e', vs.Object(SimpleObject)),
                    ]))])).values()),
        list(
            class_schema.create_schema([
                # Order matters!
                ('b', vs.Bool().noneable()),
                ('c', vs.Dict([
                    ('e', vs.Object(SimpleObject)),
                    ('d', vs.Str()),
                ])),
                ('a', vs.Int()),
            ]).values()))

    # Override field with default value.
    self.assertEqual(
        list(
            class_schema.create_schema([
                ('b', True),
            ]).extend(class_schema.create_schema([
                ('b', vs.Bool().noneable()),
            ])).values()),
        list(class_schema.create_schema([
            ('b', vs.Bool(True)),
        ]).values()))

    # Extend a base schema with non-const keys while current schema does not
    # allow const keys.
    with self.assertRaisesRegex(
        ValueError,
        'Non-const key .* is not allowed to be added to the schema'):
      class_schema.create_schema(
          [('b', vs.Int())],
          allow_nonconst_keys=False
          ).extend(
              class_schema.create_schema([
                  (ks.StrKey(), vs.Int())
              ], allow_nonconst_keys=True))

    # Override field with different type.
    with self.assertRaisesRegex(
        TypeError, 'Int\\(default=1\\) cannot extend Bool\\(.*\\): '
        'incompatible type. \\(path=b\\)'):
      class_schema.create_schema([
          ('b', 1),
      ]).extend(class_schema.create_schema([
          ('b', vs.Bool().noneable()),
      ]))

    # Override metadata.
    self.assertEqual(
        class_schema.create_schema([], metadata={
            'meta2': 'bar',
            'meta3': 'new'
        }).extend(
            class_schema.create_schema([], metadata={
                'meta1': 1,
                'meta2': 'foo'
            })).metadata, {
                'meta1': 1,
                'meta2': 'bar',
                'meta3': 'new'
            })

  def test_is_compatible(self):
    with self.assertRaisesRegex(
        TypeError, 'Argument \'other\' should be a Schema object'):
      class_schema.create_schema([]).is_compatible(1)

    self.assertFalse(
        class_schema.create_schema([]).is_compatible(
            class_schema.create_schema([('a', vs.Int())])))

    self.assertFalse(
        class_schema.create_schema([
            ('a', vs.Int())
        ]).is_compatible(class_schema.create_schema([])))

    self.assertFalse(
        class_schema.create_schema([
            ('a', vs.Int())
        ]).is_compatible(class_schema.create_schema([('a', vs.Str())])))

    self.assertTrue(
        class_schema.create_schema([
            ('a', vs.Any())
        ]).is_compatible(class_schema.create_schema([('a', vs.Str())])))

  def test_get_field(self):
    s = class_schema.create_schema([
        ('a', vs.Int()),
        (ks.StrKey('foo.*'), vs.Int()),
        (ks.StrKey('f.*'), vs.Bool())
    ], allow_nonconst_keys=True)
    self.assertEqual(s.get_field('a'), Field('a', vs.Int()))
    self.assertIsNone(s.get_field('b'))
    self.assertEqual(
        s.get_field('foo1'), Field(ks.StrKey('foo.*'), vs.Int()))
    self.assertEqual(
        s.get_field('far'), Field(ks.StrKey('f.*'), vs.Bool()))

  def test_resolve(self):
    s = class_schema.create_schema([
        ('a', vs.Int()),
        ('b', vs.Int()),
        (ks.StrKey('foo.*'), vs.Int()),
    ], allow_nonconst_keys=True)
    matched, unmatched = s.resolve(['a', 'b', 'c', 'foo1', 'foo2', 'd'])
    self.assertEqual(matched, {
        'a': ['a'],
        'b': ['b'],
        ks.StrKey('foo.*'): ['foo1', 'foo2'],
    })
    self.assertEqual(unmatched, ['c', 'd'])

  def test_dynamic_field(self):
    # No dynamic field.
    s = class_schema.create_schema([
        ('b', vs.Bool())
    ], allow_nonconst_keys=True)
    self.assertIsNone(s.dynamic_field)

    # Self-declared dynamic field.
    s = class_schema.create_schema([
        ('a', vs.Int()),
        (ks.StrKey(), vs.Str())
    ], allow_nonconst_keys=True)
    self.assertEqual(s.dynamic_field, Field(ks.StrKey(), vs.Str()))

    # Inherited dynamic field.
    s = class_schema.create_schema([
        ('b', vs.Bool())
    ], base_schema_list=[s], allow_nonconst_keys=True)
    self.assertEqual(s.dynamic_field, Field(ks.StrKey(), vs.Str()))

  def test_Validate(self):
    # Validate fully specified fields.
    self._create_test_schema().validate({
        'a': 1,
        'b': False,
        'c': {
            'd': [None],
            'e': [{
                'foo1': 'bar'
            }],
            'f': SimpleObject()
        }
    })

    # Validate partially specified fields.
    # This should pass since field 'c' is noneable.
    self._create_test_schema().validate({})

    # Missing required field 'c.e'.
    with self.assertRaisesRegex(
        ValueError, 'Required value is not specified. \\(Path=\'c\\.e\'.*\\)'):
      self._create_test_schema().validate({'c': {}})

    # Wrong key in field 'c.e[0]'
    with self.assertRaisesRegex(
        KeyError, 'Keys \\[\'bar\'\\] are not allowed in Schema. \\(parent='
        '\'c\\.e\\[0\\]\'\\)'):
      self._create_test_schema().validate({'c': {'e': [{'bar': 'foo'}]}})

    # Wrong type in field 'c.f'
    with self.assertRaisesRegex(
        TypeError, 'Expect <class \'pyglove.core.typing.class_schema_test.SimpleObject\'> but encountered '
        '<(type|class) \'int\'>: 1. \\(path=c\\.f\\)'):
      self._create_test_schema().validate(
          {'c': {
              'e': [{
                  'foo': 'bar'
              }],
              'f': 1
          }})

  def test_apply(self):
    """Tests for Schema.apply."""
    # Use default.
    self.assertEqual(
        self._create_test_schema().apply({}), {
            'a': 1,
            'b': None,
            'c': None,
        })

    # Apply using complex object.
    so = SimpleObject()
    self.assertEqual(
        self._create_test_schema().apply({
            'a': 2,
            'b': True,
            'c': {
                'e': [{
                    'foo': 'bar'
                }],
                'f': so
            }
        }), {
            'a': 2,
            'b': True,
            'c': {
                'd': [0, 1],
                'e': [{
                    'foo': 'bar'
                }],
                'f': so
            }
        })

    # Raises when input doesn't match schema.
    with self.assertRaisesRegex(
        ValueError, 'Required value is not specified. \\(Path=\'c\\.e\'.*\\)'):
      self._create_test_schema().apply({'c': {}})

    self.assertEqual(
        self._create_test_schema().apply({'c': {}}, allow_partial=True),
        {
            'a': 1,
            'b': None,
            'c': {
                'd': [0, 1],
                # NOTE: we can use typed_missing.MISSING_VALUE as general
                # missing value identifier.
                'e': typed_missing.MISSING_VALUE,
                # Or we can use missing value for specific value spec.
                'f': typed_missing.MissingValue(vs.Object(SimpleObject)),
            }
        })

  def test_apply_with_custom_typing(self):

    class NumberType(custom_typing.CustomTyping):

      def __init__(self, value):
        self.value = value

      def custom_apply(self, root_path, value_spec, **kwargs):
        # Pass through value to Enum.apply.
        if isinstance(value_spec, vs.Enum):
          return (True, self.value)
        if not isinstance(value_spec, vs.Number):
          raise ValueError(
              f'NumberType can only apply to numbers. (path=\'{root_path}\')')
        return (False, self)

      def __eq__(self, other):
        return isinstance(other, NumberType) and self.value == other.value

      def __ne__(self, other):
        return not self.__eq__(other)

    s = class_schema.create_schema([
        ('a', vs.Int(1)),
        ('b', vs.Bool().noneable()),
        ('c', vs.Dict([
            ('d', vs.Int()),
            ('e', vs.Enum(0, [0, 1])),
        ]).noneable())])

    self.assertEqual(
        s.apply({
            'a': NumberType(1),
            'c': {
                'd': NumberType(2),
                'e': NumberType(0),
            }
        }), {
            'a': NumberType(1),
            'b': None,
            'c': {
                'd': NumberType(2),
                'e': 0
            }
        })

    with self.assertRaisesRegex(
        ValueError, 'NumberType can only apply to numbers.'):
      s.apply({
          'b': NumberType(1),
      })

  def test_apply_custom_transform(self):
    """Tests for Schema.apply with custom transform_fn."""
    transformed = []

    def _transform_fn(path, field, value):
      transformed.append((path, field.key, copy.deepcopy(value)))
      if isinstance(field.value, vs.Int):
        return value + 1
      return value

    self.assertEqual(
        self._create_test_schema().apply(
            {
                'b': False,
                'c': {
                    'd': [0],
                    'e': [{
                        'foo': 'bar'
                    }],
                },
            },
            allow_partial=True,
            child_transform=_transform_fn),
        {
            'a': 2,
            'b': False,
            'c': {
                # d is Enum type.
                'd': [0],
                'e': [{
                    'foo': 'bar'
                }],
                'f': typed_missing.MISSING_VALUE,
            },
        })

    self.assertEqual(transformed, [
        ('a', 'a', 1),
        ('b', 'b', False),
        ('c.d[0]', ks.ListKey(), 0),
        ('c.d', 'd', [0]),
        ('c.e[0].foo', ks.StrKey('foo.*'), 'bar'),
        ('c.e[0]', ks.ListKey(), {
            'foo': 'bar'
        }),
        ('c.e', 'e', [{
            'foo': 'bar'
        }]),
        ('c.f', 'f', typed_missing.MISSING_VALUE),
        ('c', 'c', {
            'd': [0],
            'e': [{
                'foo': 'bar'
            }],
            'f': typed_missing.MISSING_VALUE
        }),
    ])

  def test_default_values(self):
    self.assertEqual(
        self._create_test_schema().apply({}, allow_partial=False),
        {
            'a': 1,
            'b': None,
            'c': None,
        })

    self.assertEqual(
        self._create_test_schema().apply({'c': {}}, allow_partial=True),
        {
            'a': 1,
            'b': None,
            'c': {
                'd': [0, 1],
                'e': typed_missing.MISSING_VALUE,
                'f': typed_missing.MISSING_VALUE,
            }
        })


class CreateSchemaTest(unittest.TestCase):
  """Tests for class_schema.create_schema."""

  def test_basics(self):
    s = class_schema.create_schema(
        [
            ('a', 1),
            ('b', 'foo', 'field b'),
            ('c', True, 'field c', {'user_data': 1}),
            ('d', 1.0),
            ('e', vs.Enum(0, [0, 1])),
            ('f', int),
            ('f1', Optional[int]),
            ('g', float),
            ('g1', Optional[float]),
            ('h', bool),
            ('h1', Optional[bool]),
            ('i', str),
            ('i1', Optional[str]),
            ('j', Union[int, float, bool]),
            ('k', list),
            ('l', list[int]),
            ('L', List[int]),
            ('l1', Optional[list[int]]),
            ('l2', ['black', 'white']),
            ('m', {'brand': 'google'}),
            ('n', None),
        ],
        'schema1',
        metadata={'user_data': 2},
    )

    self.assertEqual(s.name, 'schema1')
    self.assertEqual(s['a'], Field('a', vs.Int(1)))
    self.assertEqual(s['b'],
                     Field('b', vs.Str('foo'), 'field b'))
    self.assertEqual(
        s['c'],
        Field('c', vs.Bool(True), 'field c', {'user_data': 1}))
    self.assertEqual(s['d'], Field('d', vs.Float(1.0)))
    self.assertEqual(s['e'], Field('e', vs.Enum(0, [0, 1])))
    self.assertEqual(s.metadata, {'user_data': 2})
    self.assertEqual(s['f'], Field('f', vs.Int()))
    self.assertEqual(s['f1'], Field('f1', vs.Int().noneable()))
    self.assertEqual(s['g'], Field('g', vs.Float()))
    self.assertEqual(s['g1'], Field('g1', vs.Float().noneable()))
    self.assertEqual(s['h'], Field('h', vs.Bool()))
    self.assertEqual(s['h1'], Field('h1', vs.Bool().noneable()))
    self.assertEqual(s['i'], Field('i', vs.Str()))
    self.assertEqual(s['i1'], Field('i1', vs.Str().noneable()))
    self.assertEqual(
        s['j'], Field('j', vs.Union([vs.Int(), vs.Float(), vs.Bool()]))
    )
    self.assertEqual(s['k'], Field('k', vs.List(vs.Any())))
    self.assertEqual(s['l'], Field('l', vs.List(vs.Int())))
    self.assertEqual(s['L'], Field('L', vs.List(vs.Int())))
    self.assertEqual(s['l1'], Field('l1', vs.List(vs.Int()).noneable()))
    self.assertEqual(
        s['l2'],
        Field('l2', vs.List(vs.Str()).set_default(['black', 'white'])),
    )
    self.assertEqual(
        s['m'], Field('m', vs.Dict().set_default({'brand': 'google'}))
    )
    self.assertEqual(s['n'], Field('n', vs.Any().noneable()))

  def test_bad_cases(self):
    with self.assertRaisesRegex(
        TypeError, 'Metadata of schema should be a dict.'
    ):
      class_schema.create_schema([], metadata=1)

    with self.assertRaisesRegex(
        TypeError, 'Field definition should be tuples with 2 to 4 elements.'
    ):
      class_schema.create_schema(['a'])

    with self.assertRaisesRegex(
        TypeError, 'Field definition should be tuples with 2 to 4 elements.'
    ):
      class_schema.create_schema([('a',)])


if __name__ == '__main__':
  unittest.main()



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Thread-local utilities."""

import contextlib
import threading
from typing import Any, Callable, Iterator

from pyglove.core.object_utils.missing import MISSING_VALUE

_RAISE_IF_NOT_FOUND = (MISSING_VALUE,)
_thread_local_state = threading.local()


@contextlib.contextmanager
def thread_local_value_scope(
    key: str, value_in_scope: Any, initial_value: Any
) -> Iterator[None]:
  """Context manager to set a thread local state within the scope."""
  previous_value = getattr(_thread_local_state, key, initial_value)
  try:
    setattr(_thread_local_state, key, value_in_scope)
    yield
  finally:
    setattr(_thread_local_state, key, previous_value)


def thread_local_has(key: str) -> bool:
  """Deletes thread-local value by key."""
  return hasattr(_thread_local_state, key)


def thread_local_set(key: str, value: Any) -> None:
  """Sets thread-local value by key."""
  setattr(_thread_local_state, key, value)


def thread_local_get(
    key: str, default_value: Any = _RAISE_IF_NOT_FOUND) -> Any:
  """Gets thread-local value."""
  value = getattr(_thread_local_state, key, default_value)
  if value is _RAISE_IF_NOT_FOUND:
    raise ValueError(f'Key {key!r} does not exist in thread-local storage.')
  return value


def thread_local_del(key: str) -> None:
  """Deletes thread-local value by key."""
  delattr(_thread_local_state, key)


def thread_local_map(
    key: str,
    value_fn: Callable[[Any], Any],
    default_initial_value: Any = _RAISE_IF_NOT_FOUND) -> Any:
  """Map a thread-local value."""
  value = thread_local_get(key, MISSING_VALUE)
  if value == MISSING_VALUE:
    value = default_initial_value
    if value is _RAISE_IF_NOT_FOUND:
      raise ValueError(f'Key {key!r} does not exist in thread-local storage.')
    thread_local_set(key, value)

  new_value = value_fn(value)
  if value is not new_value:
    thread_local_set(key, new_value)
  return new_value


def thread_local_increment(key: str, default_initial_value: int = 0) -> int:
  """Increment an integer identified by key."""
  return thread_local_map(
      key,
      lambda x: x + 1,
      default_initial_value=default_initial_value
  )


def thread_local_decrement(
    key: str,
    default_initial_value: int = _RAISE_IF_NOT_FOUND  # pytype: disable=annotation-type-mismatch
    ) -> int:
  """Increment an integer identified by key."""
  return thread_local_map(
      key,
      lambda x: x - 1,
      default_initial_value=default_initial_value
  )


def thread_local_push(key: str, value: Any) -> None:
  """Pushes a value to a stack identified by key."""
  thread_local_map(
      key,
      lambda x: x.append(value) or x,
      default_initial_value=[]
  )


def thread_local_pop(key: str, default_value: Any = _RAISE_IF_NOT_FOUND) -> Any:
  """Pops a value from a stack identified by key."""
  stack = thread_local_get(key, MISSING_VALUE)
  if stack == MISSING_VALUE:
    if default_value is _RAISE_IF_NOT_FOUND:
      raise ValueError(f'Key {key!r} does not exist in thread-local storage.')
    return default_value

  if not isinstance(stack, list):
    raise TypeError(
        f'Key {key!r} from thread-local storage is not a list: {stack}')

  if not stack and default_value is not _RAISE_IF_NOT_FOUND:
    return default_value
  return stack.pop()



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Hyper objects: representing template-based object space.

In PyGlove, an object space is represented by a hyper object, which is an
symbolic object that is placeheld by hyper primitives
(:class:`pg.hyper.HyperPrimitive`). Through hyper objects, object templates
(:class:`pg.hyper.ObjectTemplate`) can be obtained to generate objects
based on program genomes (:class:`pg.DNA`).

 .. graphviz::
    :align: center

    digraph hypers {
      node [shape="box"];
      edge [arrowtail="empty" arrowhead="none" dir="back" style="dashed"];
      hyper [label="HyperValue" href="hyper_value.html"];
      template [label="ObjectTemplate" href="object_template.html"];
      primitive [label="HyperPrimitive" href="hyper_primitive.html"];
      choices [label="Choices" href="choices.html"];
      oneof [label="OneOf" href="oneof_class.html"];
      manyof [label="ManyOf" href="manyof_class.html"];
      float [label="Float" href="float.html"];
      custom [label="CustomHyper" href="custom_hyper.html"];
      evolve [label="Evolvable" href="evolvable.html"];
      hyper -> template;
      hyper -> primitive;
      primitive -> choices;
      choices -> oneof;
      choices -> manyof;
      primitive -> float;
      primitive -> custom;
      custom -> evolve;
    }

Hyper values map 1:1 to genotypes as the following:

+---------------------------------+----------------------------------------+
| Hyper class                     | Genotype class                         |
+=================================+========================================+
|:class:`pg.hyper.HyperValue`     |:class:`pg.DNASpec`                     |
+---------------------------------+----------------------------------------+
|:class:`pg.hyper.ObjectTemplate` |:class:`pg.geno.Space`                  |
+---------------------------------+----------------------------------------+
|:class:`pg.hyper.HyperPrimitive` |:class:`pg.geno.DecisionPoint`          |
+---------------------------------+----------------------------------------+
|:class:`pg.hyper.Choices`        |:class:`pg.geno.Choices`                |
+---------------------------------+----------------------------------------+
|:class:`pg.hyper.Float`          |:class:`pg.geno.Float`                  |
+---------------------------------+----------------------------------------+
|:class:`pg.hyper.CustomHyper`    |:class:`pg.geno.CustomDecisionPoint`    |
+---------------------------------+----------------------------------------+
"""

# pylint: disable=g-bad-import-order

# The hyper value interface and hyper primitives.
from pyglove.core.hyper.base import HyperValue
from pyglove.core.hyper.base import HyperPrimitive

from pyglove.core.hyper.categorical import Choices
from pyglove.core.hyper.categorical import OneOf
from pyglove.core.hyper.categorical import ManyOf
from pyglove.core.hyper.numerical import Float
from pyglove.core.hyper.custom import CustomHyper

from pyglove.core.hyper.evolvable import Evolvable
from pyglove.core.hyper.evolvable import MutationType
from pyglove.core.hyper.evolvable import MutationPoint

# Helper functions for creating hyper values.
from pyglove.core.hyper.categorical import oneof
from pyglove.core.hyper.categorical import manyof
from pyglove.core.hyper.categorical import permutate
from pyglove.core.hyper.numerical import floatv
from pyglove.core.hyper.evolvable import evolve

# Object template and helper functions.
from pyglove.core.hyper.object_template import ObjectTemplate
from pyglove.core.hyper.object_template import template
from pyglove.core.hyper.object_template import materialize
from pyglove.core.hyper.object_template import dna_spec

from pyglove.core.hyper.derived import DerivedValue
from pyglove.core.hyper.derived import ValueReference
from pyglove.core.hyper.derived import reference

# Classes and functions for dynamic evaluation.
from pyglove.core.hyper.dynamic_evaluation import dynamic_evaluate
from pyglove.core.hyper.dynamic_evaluation import DynamicEvaluationContext
from pyglove.core.hyper.dynamic_evaluation import trace


# Helper functions for iterating examples from the search space.
from pyglove.core.hyper.iter import iterate
from pyglove.core.hyper.iter import random_sample


# Alias for backward compatibility:
ChoiceList = ManyOf
ChoiceValue = OneOf
Template = ObjectTemplate
one_of = oneof
sublist_of = manyof
float_value = floatv
search_space = dna_spec


# pylint: enable=g-bad-import-order



# Copyright 2019 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for pyglove.symbolic.class_wrapper."""

import copy
import dataclasses
import inspect
import unittest

from pyglove.core import detouring as pg_detouring
from pyglove.core import typing as pg_typing

from pyglove.core.symbolic.base import eq as pg_eq
from pyglove.core.symbolic.base import ne as pg_ne
from pyglove.core.symbolic.base import query as pg_query
from pyglove.core.symbolic.base import sym_hash as pg_hash
from pyglove.core.symbolic.class_wrapper import apply_wrappers as pg_apply_wrappers
from pyglove.core.symbolic.class_wrapper import ClassWrapper
from pyglove.core.symbolic.class_wrapper import wrap as pg_wrap
from pyglove.core.symbolic.class_wrapper import wrap_module as pg_wrap_module
from pyglove.core.symbolic.pure_symbolic import NonDeterministic


class WrapTest(unittest.TestCase):
  """Tests for `pg.wrap`."""

  def test_wrap_with_super_init(self):

    class A:
      def __init__(self):
        self.y = 0

    class B(A):

      def __init__(self, x):
        super().__init__()
        self.x = x

    B1 = pg_wrap(B)  # pylint: disable=invalid-name
    b = B1(1)
    self.assertIsInstance(b, B)
    self.assertIsInstance(b, ClassWrapper)
    self.assertEqual(b.x, 1)
    self.assertEqual(b.y, 0)

  def test_wrap_with_varargs_and_kwargs(self):

    class A:

      def __init__(self, x, *args, y=1, **kwargs):
        self.x = x
        self.varargs = args
        self.y = y
        self.kw = kwargs

    A1 = pg_wrap(A)  # pylint: disable=invalid-name
    a = A1(0, 1, 2, y=3, p=4, q=5)
    self.assertEqual(a.x, 0)
    self.assertEqual(a.varargs, (1, 2))
    self.assertEqual(a.y, 3)
    self.assertEqual(a.kw, {'p': 4, 'q': 5})

    self.assertEqual(a.rebind({'args[0]': 0}).varargs, (0, 2))
    self.assertEqual(
        a.rebind({'p': 5, 'q': 6, 'r': 7}).kw,
        {'p': 5, 'q': 6, 'r': 7})

  def test_wrap_with_no_typing(self):

    class A:

      def __init__(self, x, y):
        self.x = x
        self.y = y
        self.z = x + y

    A1 = pg_wrap(A)  # pylint: disable=invalid-name
    a = A1(x=1, y=1)
    self.assertIsInstance(a, ClassWrapper)
    self.assertIsInstance(a, A)
    self.assertEqual(inspect.getfullargspec(A).args, ['self', 'x', 'y'])

    # Test equality semantics doesn't change.
    self.assertNotEqual(a, A1(1, 1))
    self.assertTrue(a, a.sym_eq(A1(1, 1)))
    self.assertNotEqual(hash(a), hash(A1(1, 1)))
    self.assertEqual(pg_hash(a), pg_hash(A1(1, 1)))

    self.assertEqual(a.x, 1)
    self.assertEqual(a.z, 2)
    a.rebind(y=2)
    self.assertEqual(a.z, 3)

    class B:    # pylint: disable=unused-variable

      def __init__(unused_x):  # pylint: disable=no-self-argument
        pass

    with self.assertRaisesRegex(
        ValueError, '.*__init__ must have `self` as the first argument'):
      _ = pg_wrap(B)

  def test_wrap_with_auto_typing(self):

    class A:

      def __init__(self, x: int, y: int = 1, **kwargs):
        """Class A.

        Args:
          x: The first integer.
          y: The second integer.
          **kwargs: Other arguments.
        """
        self.z = x + y + sum(kwargs.values())

    A1 = pg_wrap(A, auto_typing=True)  # pylint: disable=invalid-name
    self.assertEqual(A1.schema.get_field('x').value, pg_typing.Int())
    self.assertEqual(A1.schema.get_field('y').value, pg_typing.Int(default=1))

  def test_wrap_with_user_typing(self):

    class B:

      def __init__(self, x, y):
        self.z = x + y

    class C(B):
      pass

    C1 = pg_wrap(C, [('x', pg_typing.Int())])  # pylint: disable=invalid-name
    self.assertEqual(inspect.getfullargspec(C1).args, ['self', 'x', 'y'])
    self.assertIn('x', C1.schema)
    self.assertIn('y', C1.schema)
    self.assertNotIn('z', C1.schema)

    c = C1(1, 2)
    self.assertIsInstance(c, ClassWrapper)
    self.assertIsInstance(c, B)
    self.assertFalse(hasattr(c, 'x'))
    self.assertFalse(hasattr(c, 'y'))

    with self.assertRaisesRegex(KeyError, 'Key .* is not allowed'):
      c.rebind(z=4)

    with self.assertRaisesRegex(TypeError, 'Expect .* but encountered .*'):
      C1(0.5, 1)

  def test_wrap_with_user_typing_for_kwargs(self):
    # When `__init__` accepts `**kwargs`, additional arugment names (e.g: `p`
    # and `q`) can be defined as symbolic arguments.

    class A:

      def __init__(self, x, y=1, **kwargs):
        self.z = x + y + sum(kwargs.values())

    A1 = pg_wrap(A, [   # pylint: disable=invalid-name
        ('p', pg_typing.Int()),
        ('q', pg_typing.Int()),
    ])
    a = A1(1, p=2, q=3)
    self.assertEqual(a.z, 7)   # 1 + 1 + 2 + 3
    self.assertNotIn('z', a.sym_init_args)

  def test_wrap_with_auto_doc(self):

    class A:
      """A test class."""

      def __init__(self, x, y):
        """Init method.

        Args:
          x: Argument x.
          y: Argument y.
        """
        self.x = x
        self.y = y

    A1 = pg_wrap(A, auto_doc=True)  # pylint: disable=invalid-name
    self.assertEqual(A1.schema.description, 'A test class.')
    self.assertEqual(list(A1.schema.fields.values()), [
        pg_typing.Field('x', pg_typing.Any(), 'Argument x.'),
        pg_typing.Field('y', pg_typing.Any(), 'Argument y.'),
    ])

  def test_wrap_dataclass_with_auto_doc(self):

    @dataclasses.dataclass
    class A:
      """A test class.

      Attributes:
        x: Argument x.
        y: Argument y.
      """
      x: int
      y: str

    A1 = pg_wrap(A, auto_doc=True)  # pylint: disable=invalid-name
    self.assertEqual(A1.schema.description, 'A test class.')
    self.assertEqual(list(A1.schema.fields.values()), [
        pg_typing.Field('x', pg_typing.Any(annotation=int), 'Argument x.'),
        pg_typing.Field('y', pg_typing.Any(annotation=str), 'Argument y.'),
    ])

  def test_automatic_reset_state(self):

    class A:

      def __init__(self, v):
        if hasattr(self, 'x'):
          self.y = v
        else:
          self.x = v

    A1 = pg_wrap(A)  # pylint: disable=invalid-name
    a = A1(1)
    self.assertEqual(a.x, 1)
    self.assertEqual(a.sym_init_args.v, 1)
    self.assertFalse(hasattr(a, 'y'))
    a.rebind(v=2)
    self.assertEqual(a.x, 2)
    self.assertEqual(a.sym_init_args.v, 2)
    self.assertFalse(hasattr(a, 'y'))

  def test_wrap_with_reset_fn(self):
    context = dict(id=0)

    class A:

      def __init__(self, x):
        self.state = {context['id']: x}
        context['id'] += 1

    def reset_state_fn(unused_obj):
      context['id'] = 0

    a1 = A(1)
    self.assertEqual(a1.state, {0: 1})
    a2 = A(2)
    self.assertEqual(a2.state, {1: 2})

    A1 = pg_wrap(A, reset_state_fn=reset_state_fn)  # pylint: disable=invalid-name
    a1 = A1(1)
    self.assertEqual(a1.state, {0: 1})
    a2 = A1(2)
    self.assertEqual(a2.state, {0: 2})

  def test_wrap_with_symbolic_repr(self):

    class A:
      pass

    A1 = pg_wrap(A)  # pylint: disable=invalid-name
    self.assertEqual(repr(A1()), 'A()')

    # When the source class does not have __repr__, symbolic
    # repr will be used even when `repr` is set to False.
    A2 = pg_wrap(A, repr=False)  # pylint: disable=invalid-name
    self.assertRegex(repr(A2()), '.* object at 0x.*')

    class B:
      def __repr__(self):
        return 'custom_repr'

    # When the source class has __repr__, symbolic
    # repr will not be used even when `repr` is set to True.
    B1 = pg_wrap(B)  # pylint: disable=invalid-name
    self.assertEqual(repr(B1()), 'B()')

    B2 = pg_wrap(B, repr=False)  # pylint: disable=invalid-name
    self.assertEqual(repr(B2()), 'custom_repr')

  def test_wrap_with_symbolic_eq(self):

    class A:
      pass

    # By default argument `eq` equals False.
    A1 = pg_wrap(A)  # pylint: disable=invalid-name
    self.assertNotEqual(A1(), A1())
    self.assertTrue(pg_eq(A1(), A1()))

    A2 = pg_wrap(A, eq=True)  # pylint: disable=invalid-name
    self.assertEqual(A2(), A2())
    self.assertTrue(pg_eq(A2(), A2()))

    class B:

      def __eq__(self, other):
        return False

    B1 = pg_wrap(B, eq=True)  # pylint: disable=invalid-name
    self.assertNotEqual(B1(), B1())

  def test_bad_wrap(self):
    with self.assertRaisesRegex(
        TypeError, 'Class wrapper can only be created from classes'):
      pg_wrap('abc')

    with self.assertRaisesRegex(
        TypeError, 'Expect .* but encountered .*'):

      class A:

        def __init__(self, x, y=1):
          self.z = x + y

      _ = pg_wrap(A, [('y', pg_typing.Bool())])

    with self.assertRaisesRegex(
        KeyError, 'found extra symbolic argument \'z\''):

      class B:

        def __init__(self, x, y=1):
          self.z = x + y

      _ = pg_wrap(B, [('z', pg_typing.Int())])


class WrapModuleTest(unittest.TestCase):
  """Tests for `pg.wrap_module`."""

  def setUp(self):
    super().setUp()

    class A:
      pass

    class B:
      pass

    class D(A):
      pass

    class Module:
      """Fake module."""

      def __init__(self, name, **kwargs):
        self.__name__ = name
        self._syms = kwargs

      def __dir__(self):
        return sorted(self._syms.keys())

      def __getattr__(self, name):
        return self._syms[name]

    self._A = A  # pylint: disable=invalid-name
    self._B = B  # pylint: disable=invalid-name
    self._D = D  # pylint: disable=invalid-name
    self._Module = Module  # pylint: disable=invalid-name
    self._src_module = Module('source_module', A=A, B=B, C=A, D=D)

  def test_wrap_all(self):
    wrapper_classes = pg_wrap_module(self._src_module)
    self.assertEqual(len(wrapper_classes), 3)
    self.assertIsInstance(wrapper_classes[0](), self._A)
    self.assertIsInstance(wrapper_classes[0](), ClassWrapper)
    self.assertIsInstance(wrapper_classes[1](), self._B)
    self.assertIsInstance(wrapper_classes[1](), ClassWrapper)
    self.assertIsInstance(wrapper_classes[2](), self._D)
    self.assertIsInstance(wrapper_classes[2](), ClassWrapper)

  def test_manual_selected_classes(self):
    wrapper_classes = pg_wrap_module(self._src_module, ['B'])
    self.assertEqual(len(wrapper_classes), 1)
    self.assertIsInstance(wrapper_classes[0](), self._B)
    self.assertIsInstance(wrapper_classes[0](), ClassWrapper)
    self.assertEqual(wrapper_classes[0].__module__, 'pyglove.core.symbolic.class_wrapper_test')

  def test_where(self):
    wrapper_classes = pg_wrap_module(
        self._src_module, where=lambda c: issubclass(c, self._A))

    self.assertEqual(len(wrapper_classes), 2)
    self.assertIsInstance(wrapper_classes[0](), self._A)
    self.assertIsInstance(wrapper_classes[0](), ClassWrapper)
    self.assertIsInstance(wrapper_classes[1](), self._D)
    self.assertIsInstance(wrapper_classes[1](), ClassWrapper)

  def test_export_to_module(self):
    target_module = self._Module('target_module')
    wrapper_classes = pg_wrap_module(
        self._src_module, ['A', 'B'], export_to=target_module)
    self.assertIs(target_module.A, wrapper_classes[0])
    self.assertIs(target_module.B, wrapper_classes[1])
    self.assertEqual(target_module.A.__module__, 'target_module')


class ApplyWrappersTest(unittest.TestCase):
  """Tests for `apply_wrappers` context manager."""

  def test_apply_wrappers(self):
    """Test `apply_wrappers`."""

    class Base:
      pass

    class A:

      def __init__(self, x):
        self.x = x

    class B(Base):
      pass

    A1 = pg_wrap(A)  # pylint: disable=invalid-name
    B1 = pg_wrap(B)  # pylint: disable=invalid-name

    # Test apply wrappers that are explicitly passed in.
    with pg_apply_wrappers([A1]):
      self.assertIsInstance(A(1), A1)
      self.assertNotIsInstance(B(), B1)
    self.assertNotIsInstance(A(1), A1)
    self.assertEqual(A(1).x, 1)

    # Test apply wrappers on all registered wrapper classes with where clause.
    with pg_apply_wrappers(where=lambda c: issubclass(c, Base)):
      self.assertNotIsInstance(A(1), A1)
      self.assertIsInstance(B(), B1)
    self.assertNotIsInstance(B(), B1)

    # Nested apply_wrappers.
    with pg_apply_wrappers([A1]):
      self.assertIsInstance(A(1), A1)
      self.assertNotIsInstance(B(), B1)
      with pg_apply_wrappers([B1]):
        self.assertIsInstance(A(1), A1)
        self.assertIsInstance(B(), B1)
      self.assertNotIsInstance(B(), B1)
    self.assertNotIsInstance(A(1), A1)

    # Test for user class with custom __new__.
    class C:

      def __new__(cls, *args, **kwargs):
        return super(C, cls).__new__(cls, *args, **kwargs)

    C1 = pg_wrap(C)  # pylint: disable=invalid-name
    with pg_apply_wrappers([C1]):
      c = C()
      self.assertIsInstance(c, C1)


class ClassWrapperTest(unittest.TestCase):
  """Tests for class wrapper."""

  def test_wrapper_for_baseless_class(self):

    class A:
      pass

    A1 = pg_wrap(A)  # pylint: disable=invalid-name
    self.assertTrue(pg_eq(A1(), A1()))
    self.assertEqual(repr(A1), f'Symbolic[{A1.sym_wrapped_cls!r}]')
    a = A1()
    self.assertIs(a.sym_wrapped, a)
    with self.assertRaisesRegex(
        ValueError, '.* takes no argument while non-empty `args` is provided'):
      _ = pg_wrap(A, [('x', pg_typing.Int())])

    class B:

      def __init__(self, x):
        self.x = x
        self.y = x + 1

    B1 = pg_wrap(B)  # pylint: disable=invalid-name
    b = B1(1)
    self.assertTrue(pg_eq(b, B1(1)))
    self.assertEqual(b.y, 2)

  def test_wrapper_for_class_with_single_inheritance(self):
    call_count = dict(a=0, c=0)

    class A:

      def __init__(self):
        self.a = 0
        call_count['a'] += 1

    class B(A):
      pass

    B1 = pg_wrap(B)  # pylint: disable=invalid-name
    b = B1()
    self.assertTrue(pg_eq(b, B1()))
    self.assertEqual(b.a, 0)
    self.assertEqual(call_count['a'], 2)

    class C(B1):

      def __init__(self, x):
        super().__init__()
        self.c = x
        call_count['c'] += 1

    c = C(1)
    self.assertTrue(pg_eq(c, C(1)))
    # Make sure parent's __init__ is invoked.
    self.assertEqual(c.a, 0)
    # Make sure self's __init__ is invoked.
    self.assertEqual(c.c, 1)
    self.assertEqual(call_count['a'], 4)
    self.assertEqual(call_count['c'], 2)

    class D(C):
      pass

    d = D(1)
    self.assertTrue(pg_eq(d, D(1)))
    self.assertTrue(pg_ne(d, C(1)))
    # Make sure parent's __init__ is invoked.
    self.assertEqual(d.a, 0)
    # Make sure parent's __init__ is invoked.
    self.assertEqual(d.c, 1)
    self.assertEqual(call_count['a'], 7)
    self.assertEqual(call_count['c'], 5)

  def test_wrapper_for_class_with_multi_inheritance(self):

    class A:

      def __init__(self, x):
        super().__init__(x)
        self.a = x

    A1 = pg_wrap(A)  # pylint: disable=invalid-name

    class B:

      def __init__(self, x):
        super().__init__(x)
        self.b = x

    B1 = pg_wrap(B)  # pylint: disable=invalid-name

    class C(A1, B1):

      def __init__(self, x):
        super().__init__(x)
        self.c = x

    c = C(1)
    self.assertEqual(c.a, 1)
    self.assertEqual(c.b, 1)
    self.assertEqual(c.c, 1)

  def test_automatic_wrap_from_inheritance(self):

    class A:

      def __init__(self, x):
        self.x = x

    A1 = pg_wrap(A)  # pylint: disable=invalid-name

    # Automatic wrapped by inheriting a class wrapper.
    class C(A1):

      def __init__(self, x, y):
        super().__init__(x)
        self.y = y

    self.assertIsInstance(C(1, 2), ClassWrapper)
    self.assertTrue(pg_eq(C(1, 2), C(1, 2)))
    self.assertEqual(list(C.schema.fields.keys()), ['x', 'y'])
    self.assertEqual(repr(C), f'<class {C.type_name!r}>')

  def test_custom_metaclass(self):

    class CustomMeta(type):

      @property
      def foo(cls):
        return 'foo'

    class A(metaclass=CustomMeta):
      pass

    A1 = pg_wrap(A)  # pylint: disable=invalid-name
    self.assertTrue(issubclass(A1, ClassWrapper))
    self.assertTrue(issubclass(A1, A))
    self.assertEqual(A1.type_name, 'pyglove.core.symbolic.class_wrapper_test.A')
    self.assertEqual(A1.schema, pg_typing.Schema([]))
    self.assertEqual(A1.foo, 'foo')
    self.assertRegex(repr(A1), r'Symbolic\[.*\]')

  def test_custom_setattr(self):

    class A:

      def __init__(self, x):
        self.x = x

      def __setattr__(self, name, value):
        super().__setattr__(f'shadow_{name}', value)

    A1 = pg_wrap(A)  # pylint: disable=invalid-name
    a = A1(1)
    self.assertEqual(a.shadow_x, 1)
    a.rebind(x=2)
    self.assertEqual(a.shadow_x, 2)

  def test_method_clashes(self):

    class A:

      def __init__(self, x):
        self.y = x

      def get_result(self):
        return self.y_y

      def __contains__(self, key):
        return key == 'y'

      def __iter__(self):
        yield 'y'

      def keys(self):
        return ['y']

      def __getattr__(self, key):
        raise AttributeError('No extra attribute allowed.')

      def __setattr__(self, key, value):
        super().__setattr__('y_' + key, value)

      def __eq__(self, other):
        return isinstance(other, int) and other == 1

      def __ne__(self, other):
        return not self.__eq__(other)

      def __hash__(self):
        return 0

      def __copy__(self):
        return 1

      def __deepcopy__(self, memo):
        return 2

      def __repr__(self):
        return 'foo'

      def __str__(self):
        return 'bar'

    A1 = pg_wrap(A, eq=True)  # pylint: disable=invalid-name
    a = A1({'p': 1})

    # Test A1's non-symbolic members are inherited from user class.
    self.assertEqual(a.get_result(), {'p': 1})
    a.m = 0
    self.assertEqual(a.y_m, 0)
    self.assertNotIn('x', a)
    self.assertEqual(a.keys(), ['y'])
    self.assertEqual(list(iter(a)), ['y'])
    with self.assertRaisesRegex(
        AttributeError, 'No extra attribute allowed'):
      _ = a.x

    # Test A1's symbolic members are inherited from ClassWrapper.
    # Since class A has defined its own __eq__, __ne__ and __hash__,
    # Even `eq` is set to True, these methods from the
    # the user class will be used.
    self.assertNotEqual(a, A1({'p': 1}))
    self.assertTrue(a.sym_eq(A1({'p': 1})))
    self.assertFalse(a.sym_eq(A1({'p': 2})))
    self.assertEqual(hash(a), 0)
    self.assertNotEqual(pg_hash(a), 0)
    self.assertEqual(copy.copy(a), 1)
    self.assertEqual(copy.deepcopy(a), 2)
    self.assertTrue(a.sym_eq(a.clone(deep=True)))
    self.assertTrue(a.sym_eq(a.clone(deep=False)))
    self.assertEqual(a.sym_init_args, dict(x=dict(p=1)))
    self.assertEqual(pg_query(A1({'p': 2}), where=lambda v: v == 2),
                     {'x.p': 2})
    self.assertEqual(str(a), inspect.cleandoc('''
        A(
          x = {
            p = 1
          }
        )
        '''))
    self.assertEqual(repr(a), 'A(x={p=1})')

    A2 = pg_wrap(  # pylint: disable=invalid-name
        A,
        repr=False,
        override=dict(get_result=lambda self: 'overridden results'))
    a = A2({'p': 1})

    # Test get_results is overriden.
    self.assertEqual(a.get_result(), 'overridden results')

    # Test A2's symbolic members are not inherited from ClassWrapper.
    self.assertEqual(a, 1)
    self.assertFalse(pg_eq(a, 1))
    self.assertTrue(pg_eq(a, A2({'p': 1})))
    self.assertNotEqual(a, 2)
    self.assertTrue(pg_ne(a, None))
    self.assertFalse(pg_ne(a, A2({'p': 1})))
    self.assertEqual(hash(a), 0)
    self.assertEqual(repr(a), 'foo')
    self.assertEqual(str(a), 'bar')
    self.assertEqual(copy.copy(a), 1)
    self.assertEqual(copy.deepcopy(a), 2)
    self.assertTrue(pg_eq(a, a.clone(deep=True)))
    self.assertTrue(pg_eq(a, a.clone(deep=False)))

  def test_delayed_init_if_nondeterministic(self):

    class A(NonDeterministic):
      pass

    class B:

      def __init__(self, a):
        self._a = a

    B1 = pg_wrap(B)  # pylint: disable=invalid-name
    # `B.__init__` shall not be called when it is partial.
    b = B1.partial()
    self.assertFalse(b.wrapped_cls_initialized)

    # `B.__init__` shall not be called when any of its
    # arguments has non-deterministic values.
    b = B1(A())
    self.assertFalse(b.wrapped_cls_initialized)

    # After rebinding `b.a` to a fixed value, `B.__init__`
    # will be automatically triggered.
    b.rebind(a=1)
    self.assertTrue(b.wrapped_cls_initialized)

    # When `b.a` is rebound to a non-deterministic value again,
    # `b` should be regarded as uninitialized again.
    b.rebind(a=A())
    self.assertFalse(b.wrapped_cls_initialized)

  def test_detour_with_classwrapper(self):

    class A:

      def __init__(self, x):
        self.x = x

    A1 = pg_wrap(A)    # pylint: disable=invalid-name

    def fun(unused_cls, x):
      return A1(x + 1)

    with pg_detouring.detour([(A, fun)]):
      a = A(1)
    self.assertIsInstance(a, A1)

    # Since A1 is a subclass of A, A1.__init__ will be called with 1 by
    # the Python runtime after `fun` is called, however it will not override
    # existing value which is 2.
    self.assertEqual(a.x, 2)


if __name__ == '__main__':
  unittest.main()



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Global, thread-local and scoped flags for handling symbolic objects."""

from typing import Any, Callable, ContextManager, Optional
from pyglove.core.object_utils import thread_local


#
# Global flags.
#

_ALLOW_EMPTY_FIELD_DESCRIPTION = True
_ALLOW_REPEATED_CLASS_REGISTRATION = True
_ORIGIN_STACKTRACE_LIMIT = 10

_LOAD_HANDLER = None
_SAVE_HANDLER = None


def allow_empty_field_description(allow: bool = True) -> None:
  """Allow empty field description, which is useful for testing purposes."""
  global _ALLOW_EMPTY_FIELD_DESCRIPTION
  _ALLOW_EMPTY_FIELD_DESCRIPTION = allow


def is_empty_field_description_allowed() -> bool:
  """Returns True if empty field description is allowed."""
  return _ALLOW_EMPTY_FIELD_DESCRIPTION


def allow_repeated_class_registration(allow: bool = True) -> None:
  """Allow repeated class registration, which is useful for testing purposes."""
  global _ALLOW_REPEATED_CLASS_REGISTRATION
  _ALLOW_REPEATED_CLASS_REGISTRATION = allow


def is_repeated_class_registration_allowed() -> bool:
  """Returns True if repeated class registration is allowed."""
  return _ALLOW_REPEATED_CLASS_REGISTRATION


def set_origin_stacktrace_limit(limit: int) -> None:
  """Set stack trace limit for origin tracking."""
  global _ORIGIN_STACKTRACE_LIMIT
  _ORIGIN_STACKTRACE_LIMIT = limit


def get_origin_stacktrace_limit() -> int:
  """Returns the limited depth of stacktrace for tracking."""
  return _ORIGIN_STACKTRACE_LIMIT


def set_load_handler(
    load_handler: Optional[Callable[..., Any]]) -> Optional[Callable[..., Any]]:
  """Sets global load handler.

  Args:
    load_handler: A callable object that takes arbitrary arguments and returns
      a value. `symbolic.load` method will pass through all arguments to this
      handler and return its return value.

  Returns:
    Previous global load handler.
  """
  if load_handler and not callable(load_handler):
    raise ValueError('`load_handler` must be callable.')
  global _LOAD_HANDLER
  old_handler = _LOAD_HANDLER
  _LOAD_HANDLER = load_handler
  return old_handler


def get_load_handler() -> Optional[Callable[..., Any]]:
  """Returns global load handler."""
  return _LOAD_HANDLER


def set_save_handler(
    save_handler: Optional[Callable[..., Any]]) -> Optional[Callable[..., Any]]:
  """Sets global save handler.

  Args:
    save_handler: A callable object that takes at least one argument as value to
      save. `symbolic.save` method will pass through all arguments to this
      handler and return its return value.

  Returns:
    Previous global save handler.
  """
  if save_handler and not callable(save_handler):
    raise ValueError('`save_handler` must be callable.')
  global _SAVE_HANDLER
  old_handler = _SAVE_HANDLER
  _SAVE_HANDLER = save_handler
  return old_handler


def get_save_handler() -> Optional[Callable[..., Any]]:
  """Returns global save handler."""
  return _SAVE_HANDLER


#
# Scoped flags.
#

_TLS_ENABLE_CHANGE_NOTIFICATION = '_enable_change_notification'
_TLS_ENABLE_TYPE_CHECK = '_enable_type_check'
_TLS_ENABLE_ORIGIN_TRACKING = '_enable_origin_tracking'
_TLS_ACCESSOR_WRITABLE = '_accessor_writable'
_TLS_ALLOW_PARTIAL = '_allow_partial'
_TLS_SEALED = '_sealed'
_TLS_AUTO_CALL_FUNCTORS = '_allow_auto_call_functors'


def notify_on_change(enabled: bool = True) -> ContextManager[None]:
  """Returns a context manager to enable or disable notification upon change.

  `notify_on_change` is thread-safe and can be nested. For example, in the
  following code, `_on_change` (thus `_on_bound`) method of `a` will be
  triggered due to the rebind in the inner `with` statement, and those of `b`
  will not be triggered as the outer `with` statement disables the
  notification::

    with pg.notify_on_change(False):
      with pg.notify_on_change(True):
        a.rebind(b=1)
      b.rebind(x=2)

  Args:
    enabled: If True, enable change notification in current scope.
      Otherwise, disable notification.

  Returns:
    A context manager for allowing/disallowing change notification in scope.
  """
  return thread_local.thread_local_value_scope(
      _TLS_ENABLE_CHANGE_NOTIFICATION, enabled, True
  )


def is_change_notification_enabled() -> bool:
  """Returns True if change notification is enabled."""
  return thread_local.thread_local_get(
      _TLS_ENABLE_CHANGE_NOTIFICATION, True
  )


def track_origin(enabled: bool = True) -> ContextManager[None]:
  """Returns a context manager to enable or disable origin tracking.

  `track_origin` is thread-safe and can be nested. For example::

    a = pg.Dict(x=1)
    with pg.track_origin(False):
      with pg.track_origin(True):
        # b's origin will be tracked, which can be accessed by `b.sym_origin`.
        b = a.clone()
      # c's origin will not be tracked, `c.sym_origin` returns None.
      c = a.clone()

  Args:
    enabled: If True, the origin of symbolic values will be tracked during
      object cloning and retuning from functors under current scope.

  Returns:
    A context manager for enable or disable origin tracking.
  """
  return thread_local.thread_local_value_scope(
      _TLS_ENABLE_ORIGIN_TRACKING, enabled, False
  )


def is_tracking_origin() -> bool:
  """Returns if origin of symbolic object are being tracked."""
  return thread_local.thread_local_get(_TLS_ENABLE_ORIGIN_TRACKING, False)


def enable_type_check(enabled: bool = True) -> ContextManager[None]:
  """Returns a context manager to enable or disable runtime type check.

  `enable_type_check` is thread-safe and can be nested. For example,
  in the following code, runtime type check with be `a` but not on `b`::

    with pg.enable_type_check(False):
      with pg.enable_type_check(True):
        a = pg.Dict(x=1, value_spec=pg.typing.Dict([('x', pg.typing.Int())]))
      b = pg.Dict(y=1, value_spec=pg.typing.Dict([('x', pg.typing.Int())]))


  Args:
    enabled: If True, enable runtime type check in current scope.
      Otherwise, disable runtime type check.

  Returns:
    A context manager for allowing/disallowing runtime type check.
  """
  return thread_local.thread_local_value_scope(
      _TLS_ENABLE_TYPE_CHECK, enabled, True
  )


def is_type_check_enabled() -> bool:
  """Returns True if runtme type check is enabled."""
  return thread_local.thread_local_get(_TLS_ENABLE_TYPE_CHECK, True)


def allow_writable_accessors(
    writable: Optional[bool] = True) -> ContextManager[None]:
  """Returns a context manager that makes accessor writable in scope.

  This function is thread-safe and can be nested. In the nested use case, the
  writable flag of immediate parent context is effective.

  Example::

    sd1 = pg.Dict()
    sd2 = pg.Dict(accessor_writable=False)
    with pg.allow_writable_accessors(False):
      sd1.a = 2  # NOT OK
      sd2.a = 2  # NOT OK
      with pg.allow_writable_accessors(True):
        sd1.a = 2   # OK
        sd2.a = 2  # OK
        with pg.allow_writable_accessors(None):
          sd1.a = 1  # OK
          sd2.a = 1  # NOT OK

  Args:
    writable: If True, allow write access with accessors (__setattr__,
      __setitem__) for all symbolic values in scope.
      If False, disallow write access via accessors for all symbolic values
      in scope, even if individual objects allow so.
      If None, honor object-level `accessor_writable` flag.

  Returns:
    A context manager that allows/disallows writable accessors of all
      symbolic values in scope. After leaving the scope, the
      `accessor_writable` flag of individual objects will remain intact.
  """
  return thread_local.thread_local_value_scope(
      _TLS_ACCESSOR_WRITABLE, writable, None
  )


def is_under_accessor_writable_scope() -> Optional[bool]:
  """Return True if symbolic values are treated as sealed in current context."""
  return thread_local.thread_local_get(_TLS_ACCESSOR_WRITABLE, None)


def as_sealed(sealed: Optional[bool] = True) -> ContextManager[None]:
  """Returns a context manager to treat symbolic values as sealed/unsealed.

  While the user can use `Symbolic.seal` to seal or unseal an individual object.
  This context manager is useful to create a readonly zone for operations on
  all existing symbolic objects.

  This function is thread-safe and can be nested. In the nested use case, the
  sealed flag of immediate parent context is effective.

  Example::

    sd1 = pg.Dict()
    sd2 = pg.Dict().seal()

    with pg.as_sealed(True):
      sd1.a = 2  # NOT OK
      sd2.a = 2  # NOT OK
      with pg.as_sealed(False):
        sd1.a = 2   # OK
        sd2.a = 2  # OK
        with pg.as_sealed(None):
          sd1.a = 1  # OK
          sd2.a = 1  # NOT OK

  Args:
    sealed: If True, treats all symbolic values as sealed in scope.
      If False, treats all as unsealed.
      If None, honor object-level `sealed` state.

  Returns:
    A context manager that treats all symbolic values as sealed/unsealed
      in scope. After leaving the scope, the sealed state of individual objects
      will remain intact.
  """
  return thread_local.thread_local_value_scope(_TLS_SEALED, sealed, None)


def is_under_sealed_scope() -> Optional[bool]:
  """Return True if symbolic values are treated as sealed in current context."""
  return thread_local.thread_local_get(_TLS_SEALED, None)


def allow_partial(allow: Optional[bool] = True) -> ContextManager[None]:
  """Returns a context manager that allows partial values in scope.

  This function is thread-safe and can be nested. In the nested use case, the
  allow flag of immediate parent context is effective.

  Example::

    @pg.members([
        ('x', pg.typing.Int()),
        ('y', pg.typing.Int())
    ])
    class A(pg.Object):
      pass

    with pg.allow_partial(True):
      a = A(x=1)  # Missing `y`, but OK
      with pg.allow_partial(False):
        a.rebind(x=pg.MISSING_VALUE)  # NOT OK
      a.rebind(x=pg.MISSING_VALUE)  # OK

  Args:
    allow: If True, allow partial symbolic values in scope.
      If False, do not allow partial symbolic values in scope even if
      individual objects allow so. If None, honor object-level
      `allow_partial` property.

  Returns:
    A context manager that allows/disallow partial symbolic values in scope.
      After leaving the scope, the `allow_partial` state of individual objects
      will remain intact.
  """
  return thread_local.thread_local_value_scope(_TLS_ALLOW_PARTIAL, allow, None)


def is_under_partial_scope() -> Optional[bool]:
  """Return True if partial value is allowed in current context."""
  return thread_local.thread_local_get(_TLS_ALLOW_PARTIAL, None)


def auto_call_functors(enabled: bool = True) -> ContextManager[None]:
  """Returns a context manager to enable or disable auto call for functors.

  `auto_call_functors` is thread-safe and can be nested. For example::

    @pg.symbolize
    def foo(x, y):
      return x + y

    with pg.auto_call_functors(True):
      a = foo(1, 2)
      assert a == 3
      with pg.auto_call_functors(False):
        b = foo(1, 2)
        assert isinstance(b, foo)

  Args:
    enabled: If True, enable auto call for functors.
      Otherwise, auto call will be disabled.

  Returns:
    A context manager for enabling/disabling auto call for functors.
  """
  return thread_local.thread_local_value_scope(
      _TLS_AUTO_CALL_FUNCTORS, enabled, False
  )


def should_call_functors_during_init() -> Optional[bool]:
  """Return True functors should be automatically called during __init__."""
  return thread_local.thread_local_get(_TLS_AUTO_CALL_FUNCTORS, None)



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for pyglove.Functor."""

import inspect
import io
import unittest

from pyglove.core import object_utils
from pyglove.core import typing as pg_typing
from pyglove.core.symbolic import flags
from pyglove.core.symbolic.base import from_json_str as pg_from_json_str
from pyglove.core.symbolic.dict import Dict
from pyglove.core.symbolic.functor import as_functor as pg_as_functor
from pyglove.core.symbolic.functor import Functor
from pyglove.core.symbolic.functor import functor as pg_functor
from pyglove.core.symbolic.list import List
from pyglove.core.symbolic.object import members as pg_members
from pyglove.core.symbolic.object import Object


MISSING_VALUE = object_utils.MISSING_VALUE


class FunctorTest(unittest.TestCase):
  """Tests for `pg.Functor`."""

  def test_basics(self):
    @pg_functor
    def f(a, *args, b, c=0, **kwargs):
      return a + sum(args)  + b + c + sum(kwargs.values())

    # Basic properties of a symbolic function (functor).
    x = f(1, 2, 3, b=4, c=5, d=6)
    self.assertIsInstance(x, Functor)
    self.assertEqual(x(), 1 + 2 + 3 + 4 + 5 + 6)

    self.assertEqual(x.a, 1)
    self.assertEqual(x.args, [2, 3])
    self.assertEqual(x.b, 4)
    self.assertEqual(x.c, 5)
    self.assertEqual(x.d, 6)    # kwargs.

    self.assertEqual(x.specified_args, {'a', 'args', 'b', 'c', 'd'})
    self.assertEqual(x.bound_args, {'a', 'args', 'b', 'c', 'd'})
    self.assertEqual(x.unbound_args, set())
    self.assertEqual(x.non_default_args, {'a', 'args', 'b', 'c', 'd'})
    self.assertEqual(x.default_args, set())

    # Partial binding.
    x = f.partial(b=2)
    self.assertEqual(x(1), 1 + 2 + 0)

    # Rebinding.
    x.rebind(b=3)
    self.assertEqual(x(1), 1 + 3 + 0)

    # Incremental binding.
    x.a = 1
    self.assertEqual(x(), 1 + 3 + 0)

    # Temporarily override binding at call time.
    self.assertEqual(x(a=2, override_args=True), 2 + 3 + 0)
    self.assertEqual(x.a, 1)
    self.assertEqual(x(), 1 + 3 + 0)

    # Inspection.
    self.assertEqual(repr(x), 'f(a=1, args=[], b=3, c=0)')
    self.assertEqual(
        x.format(compact=True, hide_default_values=True), 'f(a=1, b=3)')

    # Comparison.
    self.assertEqual(x, f(1, b=3))
    self.assertTrue(x.sym_lt(f(1, b=4)))

    # Serialization.
    self.assertEqual(pg_from_json_str(x.to_json_str()), x)

  def test_no_typing(self):
    @pg_functor()
    def f(a, b, *args, c=0, **kwargs):
      return a + b + c + sum(args) + sum(kwargs.values())

    self.assertEqual(
        list(f.schema.values()), [
            pg_typing.Field('a', pg_typing.Any()),
            pg_typing.Field('b', pg_typing.Any()),
            pg_typing.Field('args',
                            pg_typing.List(pg_typing.Any(), default=[])),
            pg_typing.Field('c', pg_typing.Any(default=0)),
            pg_typing.Field(pg_typing.StrKey(), pg_typing.Any()),
        ])
    self.assertEqual(f.signature.args, [
        pg_typing.Argument('a', pg_typing.Any()),
        pg_typing.Argument('b', pg_typing.Any())
    ])
    self.assertEqual(
        f.signature.varargs,
        pg_typing.Argument('args', pg_typing.Any()))
    self.assertEqual(
        f.signature.varkw, pg_typing.Argument('kwargs', pg_typing.Any()))
    self.assertEqual(
        f.signature.kwonlyargs,
        [pg_typing.Argument('c', pg_typing.Any(default=0))])
    self.assertIsNone(f.signature.return_value, None)
    self.assertTrue(f.signature.has_varargs)
    self.assertIsInstance(f.partial(), Functor)
    self.assertEqual(f.partial()(1, 2), 3)
    self.assertEqual(f.partial(b=1)(1), 2)
    self.assertEqual(f.partial(b=1)(a=1), 2)

    self.assertEqual(f(1, 2, 3, 4)(), 10)
    self.assertEqual(f(1, 2, 3, 4, c=5)(), 15)
    self.assertEqual(f(1, 2, 3, 4, c=5, x=5)(), 20)
    self.assertEqual(f(1, 2, 3, 4)(2, 3, 4, override_args=True), 9)
    self.assertEqual(f(1, 2, 3, 4)(c=1), 11)
    self.assertEqual(f(1, 2, 3, 4)(x=2), 12)
    self.assertEqual(f.partial(b=1)(2, c=4), 7)

  def test_full_typing(self):
    @pg_functor([
        ('a', pg_typing.Int()),
        ('b', pg_typing.Int()),
    ], returns=pg_typing.Int())
    def f(a=1, b=2):
      return a + b

    self.assertEqual(f.signature.args, [
        pg_typing.Argument('a', pg_typing.Int(default=1)),
        pg_typing.Argument('b', pg_typing.Int(default=2)),
    ])
    self.assertEqual(
        list(f.schema.values()), [
            pg_typing.Field('a', pg_typing.Int(default=1)),
            pg_typing.Field('b', pg_typing.Int(default=2)),
        ])
    self.assertEqual(f.signature.return_value, pg_typing.Int())
    self.assertFalse(f.signature.has_varargs)
    self.assertFalse(f.signature.has_varkw)
    self.assertEqual(f.partial()(), 3)
    self.assertEqual(f.partial(a=2)(b=2), 4)
    self.assertEqual(f.partial(a=3, b=2)(), 5)
    self.assertEqual(f.partial(1, 2)(), 3)

    # Override default value.
    self.assertEqual(f.partial(a=2)(), 4)
    self.assertEqual(f.partial(a=2, override_args=True)(a=3), 5)
    self.assertEqual(f.partial(a=1, b=1, override_args=True)(a=2, b=2), 4)
    self.assertEqual(f.partial(2, 4, override_args=True)(1), 5)

  def test_auto_typing(self):
    @pg_functor(auto_typing=True)
    def f(a: int, *args, b: int = 2, **kwargs) -> int:
      del args, kwargs
      return a + b

    self.assertEqual(
        list(f.schema.values()), [
            pg_typing.Field('a', pg_typing.Int()),
            pg_typing.Field(
                'args', 
                pg_typing.List(pg_typing.Any(), default=[])),
            pg_typing.Field('b', pg_typing.Int(default=2)),
            pg_typing.Field(pg_typing.StrKey(), pg_typing.Any()),
        ])
    self.assertEqual(
        f.signature.args,
        [pg_typing.Argument('a', pg_typing.Int())])
    self.assertEqual(
        f.signature.varargs,
        pg_typing.Argument('args', pg_typing.Any()))
    self.assertEqual(
        f.signature.kwonlyargs,
        [pg_typing.Argument('b', pg_typing.Int(default=2))])
    self.assertEqual(
        f.signature.varkw,
        pg_typing.Argument('kwargs', pg_typing.Any()))
    self.assertEqual(f.signature.return_value, pg_typing.Int())

    # Test runtime value check.
    with self.assertRaisesRegex(TypeError, 'Expect .* but encountered .*'):
      f(a='abc')

  def test_auto_doc(self):
    @pg_functor([
        ('a', pg_typing.Int()),
        ('b', pg_typing.Int()),
    ], returns=pg_typing.Int(), auto_doc=True)
    def f(a=1, b=2):
      """Compute the sum.
      
      Args:
        a: an integer.
        b: another integer.
      
      Returns:
        Sum of two integers.
      """
      return a + b

    self.assertEqual(f.schema.description, 'Compute the sum.')
    self.assertEqual(
        list(f.schema.values()), [
            pg_typing.Field('a', pg_typing.Int(default=1), 'an integer.'),
            pg_typing.Field('b', pg_typing.Int(default=2), 'another integer.'),
        ])
    self.assertEqual(f.signature.return_value, pg_typing.Int())
    self.assertFalse(f.signature.has_varargs)
    self.assertFalse(f.signature.has_varkw)
    self.assertEqual(f.partial()(), 3)
    self.assertEqual(f.partial(a=2)(b=2), 4)
    self.assertEqual(f.partial(a=3, b=2)(), 5)
    self.assertEqual(f.partial(1, 2)(), 3)

  def test_partial_typing(self):
    @pg_functor([
        ('c', pg_typing.Int()),
        ('a', pg_typing.Int()),
    ])
    def f(a, b=1, c=1):
      return a + b + c

    self.assertEqual(f.signature.args, [
        pg_typing.Argument('a', pg_typing.Int()),
        pg_typing.Argument('b', pg_typing.Any(default=1)),
        pg_typing.Argument('c', pg_typing.Int(default=1)),
    ])
    self.assertFalse(f.signature.has_varargs)
    self.assertFalse(f.signature.has_varkw)
    self.assertEqual(
        list(f.schema.values()), [
            pg_typing.Field('a', pg_typing.Int()),
            pg_typing.Field('b', pg_typing.Any(default=1)),
            pg_typing.Field('c', pg_typing.Int(default=1)),
        ])
    self.assertEqual(f.partial()(a=2), 4)
    # Override 'a' with 2, provide 'b' with 2, use 'c' from default value 1.
    self.assertEqual(f.partial()(2, 2), 5)

    # 'a' is not provided.
    with self.assertRaisesRegex(
        TypeError, 'missing 1 required positional argument'):
      f.partial()()

  def test_runtime_type_check(self):
    @pg_functor([
        ('a', pg_typing.Int(min_value=0)),
        ('b', pg_typing.Int(max_value=10)),
        ('args', pg_typing.List(pg_typing.Int())),
        (pg_typing.StrKey(), pg_typing.Int(max_value=5))
    ], returns=pg_typing.Int(min_value=0))
    def f(a, *args, b, **kwargs):
      return a + b + sum(args) + sum(kwargs.values())

    self.assertEqual(f(1, 2, b=3)(c=4), 10)

    # Validate during pre-binding.
    with self.assertRaisesRegex(
        ValueError, 'Value -1 is out of range .*min=0'):
      f(-1, b=1)

    with self.assertRaisesRegex(
        TypeError, 'Expect .*int.* but encountered .*float'):
      f(1, 0.1, b=1)

    with self.assertRaisesRegex(
        ValueError, 'Value 11 is out of range .*max=10'):
      f(1, b=11)

    with self.assertRaisesRegex(
        ValueError, 'Value 10 is out of range .*max=5'):
      f(1, b=1, c=10)

    # Validate during late binding.
    with self.assertRaisesRegex(
        ValueError, 'Value -1 is out of range .*min=0'):
      f.partial(b=2)(-1)

    with self.assertRaisesRegex(
        TypeError, 'Expect .*int.* but encountered .*float'):
      f(1, b=1)(1, 0.1, override_args=True)

    with self.assertRaisesRegex(
        ValueError, 'Value 11 is out of range .*max=10'):
      f(1)(b=11)   # pylint: disable=missing-kwoa

    with self.assertRaisesRegex(
        ValueError, 'Value 10 is out of range .*max=5'):
      f(1, b=1)(c=10)

    with self.assertRaisesRegex(
        ValueError, 'Value 6 is out of range .*max=5'):
      f(1, b=1)(c=6)

    with self.assertRaisesRegex(
        ValueError, 'Value -1 is out of range .*min=0'):
      f(0, b=-1)()

    # Returning pg.MISSING_VALUE will not trigger return value spec check.
    @pg_functor(auto_typing=True)
    def g() -> int:
      return pg_typing.MISSING_VALUE

    self.assertEqual(g()(), pg_typing.MISSING_VALUE)

  def test_symbolization_on_nested_containers(self):
    @pg_functor
    def f(a, b):
      del a, b

    x = f([], {})
    self.assertIsInstance(x.a, List)
    self.assertIsInstance(x.b, Dict)

  def test_implicit_copy_during_assignment(self):
    @pg_functor
    def f(a, b):
      del a, b

    class X:
      pass

    # There is no impliit copy when assigning a root symbolic object to
    # another tree.
    x = X()
    f1 = f(x, f({}, []))
    f2 = f(f1, 1)
    self.assertIs(f2.a, f1)

    # There is an implicit copy when assigning a symbolic object with
    # a parent to another tree.
    sd = Dict(f=f1)
    self.assertEqual(f1, sd.f)
    self.assertIsNot(f1, sd.f)
    self.assertIsNot(f1.b.a, sd.f.b.a)
    self.assertIsNot(f1.b.b, sd.f.b.b)
    # Non-symbolic member is copied by reference.
    self.assertIs(f1.a, sd.f.a)

  def test_as_functor(self):
    f = pg_as_functor(lambda x: x)
    self.assertIsInstance(f, Functor)
    self.assertEqual(
        f.signature.args, [pg_typing.Argument('x', pg_typing.Any())])
    self.assertIsNone(f.signature.return_value)
    self.assertEqual(f(1), 1)

  def test_bad_definition(self):
    # `pg.functor` decorator is not applicable to class.
    with self.assertRaisesRegex(TypeError, '.* is not a method.'):

      @pg_functor([('a', pg_typing.Int())])
      class A:
        pass

      del A

    with self.assertRaisesRegex(
        TypeError, '.* got an unexpected keyword argument'):
      @pg_functor(unsupported_keyword=1)
      def f1(x):   # pylint: disable=unused-variable
        return x

    # `pg.functor` decorator found extra symbolic argument.
    with self.assertRaisesRegex(
        KeyError, '.* found extra symbolic argument \'a\'.'):
      @pg_functor([('a', pg_typing.Int())])
      def f2():  # pylint: disable=unused-variable
        pass

    # `pg.functor` decorator found extra symbolic argument.
    with self.assertRaisesRegex(
        ValueError,
        'the value spec for positional wildcard argument .*'
        'must be a `pg.typing.List` instance'):
      @pg_functor([('args', pg_typing.Int())])
      def f3(*args):  # pylint: disable=unused-variable
        del args

    # `pg.functor` decorator has multiple StrKey.
    with self.assertRaisesRegex(
        KeyError,
        '.* multiple StrKey found in symbolic arguments declaration.'):
      @pg_functor([
          ('a', pg_typing.Int()),
          (pg_typing.StrKey(), pg_typing.Any()),
          (pg_typing.StrKey(), pg_typing.Any()),
      ])  # pylint: disable=unused-variable
      def f4(a, **kwargs):
        del a, kwargs

    with self.assertRaisesRegex(
        KeyError, '.* multiple symbolic fields found for argument.'):
      @pg_functor([
          ('a', pg_typing.Int()),
          ('a', pg_typing.Str()),
      ])  # pylint: disable=unused-variable
      def f5(a):
        del a

    with self.assertRaisesRegex(
        ValueError,
        '.* the default value .* of symbolic argument .* does not equal '
        'to the default value .* specified at function signature'):
      @pg_functor([
          ('a', pg_typing.Int(default=2)),
      ])  # pylint: disable=unused-variable
      def f6(a=1):
        del a

    with self.assertRaisesRegex(
        ValueError, 'return value spec should not have default value'):
      @pg_functor([
          ('a', pg_typing.Int()),
      ], returns=pg_typing.Any(default=None))
      def f7(a=1):  # pylint: disable=unused-variable
        del a

  def test_bad_call(self):

    @pg_functor([
        ('a', pg_typing.Int()),
    ])
    def f(a=1):
      del a

    with self.assertRaisesRegex(
        TypeError, '.* got multiple values for keyword argument \'a\''):
      f(1, a=1)  # pylint: disable=redundant-keyword-arg

    with self.assertRaisesRegex(
        TypeError, '.* takes 1 positional argument but 2 were given'):
      f(1, 2)  # pylint: disable=too-many-function-args

    with self.assertRaisesRegex(
        TypeError, '.* takes 1 positional argument but 2 were given'):
      f.partial()(1, 2)  # pylint: disable=too-many-function-args

  def test_inspect_signature(self):
    @pg_functor([
        ('a', pg_typing.Int())
    ])
    def f(a, *b, c=1, **d):
      del a, b, c, d

    signature = inspect.signature(f.__init__)
    self.assertEqual(
        list(signature.parameters.keys()), ['self', 'a', 'b', 'c', 'd'])
    self.assertEqual(
        signature.parameters['self'].kind,
        inspect.Parameter.POSITIONAL_OR_KEYWORD)
    self.assertEqual(
        signature.parameters['self'].annotation,
        inspect.Signature.empty)
    self.assertEqual(
        signature.parameters['self'].default,
        inspect.Signature.empty)

    self.assertEqual(
        signature.parameters['a'].kind,
        inspect.Parameter.POSITIONAL_OR_KEYWORD)
    self.assertEqual(signature.parameters['a'].annotation, int)
    self.assertEqual(
        signature.parameters['a'].default,
        inspect.Signature.empty)

    self.assertEqual(
        signature.parameters['b'].kind,
        inspect.Parameter.VAR_POSITIONAL)
    self.assertEqual(
        signature.parameters['b'].annotation,
        inspect.Signature.empty)

    self.assertEqual(
        signature.parameters['c'].kind,
        inspect.Parameter.KEYWORD_ONLY)
    self.assertEqual(
        signature.parameters['c'].annotation,
        inspect.Signature.empty)
    self.assertEqual(
        signature.parameters['c'].default, 1)

    self.assertEqual(
        signature.parameters['d'].kind,
        inspect.Parameter.VAR_KEYWORD)
    self.assertEqual(
        signature.parameters['d'].annotation,
        inspect.Signature.empty)

  def test_inspect(self):
    @pg_functor
    def f(x):
      del x

    s = io.StringIO()
    x = f([f(1), f(2)])
    x.inspect(file=s, compact=True)
    self.assertEqual(s.getvalue(), 'f(x=[0: f(x=1), 1: f(x=2)])\n')

    s = io.StringIO()
    x.inspect(where=lambda v: v == 1, file=s)
    self.assertEqual(s.getvalue(), '{\n  \'x[0].x\': 1\n}\n')

  def test_sym_init_args(self):

    @pg_functor
    def f(x, y=True):
      del x, y

    x = f(1)
    self.assertEqual(x.sym_init_args, {
        'x': 1,
        'y': True
    })
    x = f.partial()
    self.assertEqual(x.sym_init_args, {
        'x': MISSING_VALUE,
        'y': True
    })

  def test_setattr(self):
    @pg_functor
    def f(a, b=1):
      return a + b

    x = f(0)
    self.assertEqual(x.specified_args, set(['a']))
    self.assertEqual(x.bound_args, set(['a', 'b']))
    self.assertEqual(x.non_default_args, set(['a']))
    self.assertEqual(x.default_args, set(['b']))
    self.assertEqual(x(), 1)                          # pylint: disable=not-callable

    x.b = 2
    self.assertEqual(x.bound_args, set(['a', 'b']))
    self.assertEqual(x(), 2)                          # pylint: disable=not-callable

    x.a = 1
    self.assertEqual(x.bound_args, set(['a', 'b']))
    self.assertEqual(x(), 3)                          # pylint: disable=not-callable

    # Reset x.b to default value 1.
    x.b = MISSING_VALUE
    self.assertEqual(x.bound_args, set(['a', 'b']))
    self.assertEqual(x(), 2)                          # pylint: disable=not-callable

  def test_getattr(self):
    @pg_functor
    def f(a, b=True):
      del a, b

    x = f(1)
    self.assertEqual(x.a, 1)
    self.assertTrue(x.b)      # Use default value from the field definition.

  def test_rebind(self):
    @pg_functor
    def f(a, b=1):
      return a + b

    x = f(0)
    self.assertEqual(x.specified_args, set(['a']))
    self.assertEqual(x.bound_args, set(['a', 'b']))
    self.assertEqual(x.unbound_args, set())
    self.assertEqual(x.non_default_args, set(['a']))
    self.assertEqual(x.default_args, set(['b']))
    self.assertEqual(x(), 1)                          # pylint: disable=not-callable

    x.rebind(b=2)
    self.assertEqual(x.specified_args, set(['a', 'b']))
    self.assertEqual(x.bound_args, set(['a', 'b']))
    self.assertEqual(x.unbound_args, set())
    self.assertEqual(x.non_default_args, set(['a', 'b']))
    self.assertEqual(x.default_args, set())
    self.assertEqual(x(), 2)                          # pylint: disable=not-callable

    x.rebind(a=1)
    self.assertEqual(x.specified_args, set(['a', 'b']))
    self.assertEqual(x.bound_args, set(['a', 'b']))
    self.assertEqual(x.unbound_args, set())
    self.assertEqual(x.non_default_args, set(['a', 'b']))
    self.assertEqual(x.default_args, set())
    self.assertEqual(x(), 3)                          # pylint: disable=not-callable

    # Unspecify x.x.
    x.rebind(a=MISSING_VALUE)
    self.assertEqual(x.specified_args, set(['b']))
    self.assertEqual(x.bound_args, set(['b']))
    self.assertEqual(x.unbound_args, set(['a']))
    self.assertEqual(x.non_default_args, set(['b']))
    self.assertEqual(x.default_args, set())
    with self.assertRaisesRegex(
        TypeError, 'missing 1 required positional argument'):
      _ = x()                                          # pylint: disable=not-callable

    # Reset x.b to default value 1.
    x.rebind(b=MISSING_VALUE)
    # self.assertEqual(x.specified_args, set(['a']))
    self.assertEqual(x.bound_args, set(['b']))
    self.assertEqual(x.unbound_args, set(['a']))
    self.assertEqual(x.non_default_args, set())
    self.assertEqual(x.default_args, set(['b']))

    x.rebind(a=1)
    self.assertEqual(x(), 2)                          # pylint: disable=not-callable

  def test_partial(self):
    @pg_functor
    def f(a, b):
      del a, b

    # An incompletely bound functor is not considered partial.
    self.assertFalse(f.partial().is_partial)

    @pg_members([('x', pg_typing.Int())])
    class A(Object):
      pass

    # A bound argument with partial value will make the functor partial.
    x = f(a=A.partial(), b=1)
    self.assertTrue(x.is_partial)
    self.assertEqual(x.missing_values(), {'a.x': MISSING_VALUE})

  def test_clone(self):

    @pg_functor
    def f(a, b=1):
      return a + b

    x = f(0, override_args=True, ignore_extra_args=True).seal(True)
    # Make sure `bound_args` is correctly set while `clone` constructs
    # a new object with non-args arguments.
    # (e.g. sealed, etc.)
    y = x.clone()
    self.assertTrue(y.is_sealed)
    self.assertEqual(y.bound_args, x.bound_args)
    self.assertEqual(y.default_args, x.default_args)
    self.assertEqual(y.non_default_args, x.non_default_args)
    self.assertEqual(y._override_args, x._override_args)
    self.assertEqual(y._ignore_extra_args, x._ignore_extra_args)

    y = x.clone(deep=True)
    self.assertTrue(y.is_sealed)
    self.assertEqual(y.bound_args, x.bound_args)
    self.assertEqual(y.default_args, x.default_args)
    self.assertEqual(y.non_default_args, x.non_default_args)
    self.assertEqual(y._override_args, x._override_args)
    self.assertEqual(y._ignore_extra_args, x._ignore_extra_args)

  def test_override_args(self):
    @pg_functor()
    def f(x, *args, a=1, **kwargs):
      return x + sum(args) + a + sum(kwargs.values())

    self.assertEqual(f.partial()(1), 2)
    # Override default value is not treated as override.
    self.assertEqual(f.partial()(1, a=2), 3)
    self.assertEqual(f(1, override_args=True)(2), 3)
    self.assertEqual(f(1, override_args=True)(a=2), 3)
    with self.assertRaisesRegex(
        TypeError,
        '.* got new value for argument \'x\' from position 0'):
      f(0)(1)

    with self.assertRaisesRegex(
        TypeError,
        '.* got new value for argument \'a\' from keyword argument'):
      f(0, a=1)(a=2)

  def test_ignore_extra_args(self):
    """Test functor with/without `ignore_extra_args` flag."""
    # Test ignore extra args.
    @pg_functor()
    def f(a=1):
      return a

    self.assertEqual(f.partial(ignore_extra_args=True)(1, c=0), 1)
    self.assertEqual(f.partial()(1, c=0, ignore_extra_args=True), 1)
    self.assertEqual(f.partial()(1, 2, ignore_extra_args=True), 1)
    with self.assertRaisesRegex(
        TypeError, '.* got an unexpected keyword argument \'c\''):
      f.partial()(1, c=0)

  def test_specified_args(self):
    @pg_functor
    def f(a, *args, b=0, c=1, **kwargs):
      del a, args, b, c, kwargs

    self.assertEqual(f(0).specified_args, set(['a']))
    self.assertEqual(f(0, 1).specified_args, set(['a', 'args']))
    self.assertEqual(f(0, 1, b=0).specified_args, set(['a', 'args', 'b']))
    self.assertEqual(f(0, 1, d=0).specified_args, set(['a', 'args', 'd']))

    x = f(0, 1, b=2, c=1, d=1)
    self.assertEqual(x.specified_args, set(['a', 'args', 'b', 'c', 'd']))

    # Delete an attribute removes the bound argument.
    del x.a
    self.assertEqual(x.specified_args, set(['args', 'b', 'c', 'd']))

  @unittest.skip('need to track the original value of FieldUpdate.new_value.')
  def test_specified_args_with_removing_args_with_default(self):
    @pg_functor
    def f(a, *args, b=0, c=1, **kwargs):
      del a, args, b, c, kwargs

    x = f(0, 1, b=0, c=0, d=0)

    # Assign an attribute to MISSING_VALUE removes the bound argument.
    x.d = MISSING_VALUE
    self.assertEqual(x.specified_args, set(['a', 'args', 'c', 'd']))

    # Rebind an attribute to MISSING_VALUE also removes the bound argument.
    x.rebind(c=MISSING_VALUE)
    self.assertEqual(x.specified_args, set(['a', 'args', 'd']))

  def test_non_default_args(self):
    @pg_functor
    def f(a, *args, b=0, c=1, **kwargs):
      del a, args, b, c, kwargs

    self.assertEqual(f(0).non_default_args, set(['a']))
    self.assertEqual(f(0, 1).non_default_args, set(['a', 'args']))
    self.assertEqual(f(0, b=0, c=1).non_default_args, set(['a']))
    self.assertEqual(f(0, b=1, c=0).non_default_args, set(['a', 'b', 'c']))
    self.assertEqual(f(0, d=2).non_default_args, set(['a', 'd']))

    x = f(0)
    self.assertEqual(x.non_default_args, set(['a']))
    x.b = 1
    self.assertEqual(x.non_default_args, set(['a', 'b']))
    x.rebind(c=0)
    self.assertEqual(x.non_default_args, set(['a', 'b', 'c']))

  def test_default_args(self):
    @pg_functor
    def f(a, *args, b=0, c=1, **kwargs):
      del a, args, b, c, kwargs

    self.assertEqual(f(0).default_args, set(['args', 'b', 'c']))
    self.assertEqual(f(0, 1).default_args, set(['b', 'c']))
    self.assertEqual(f(0, b=0, c=1).default_args, set(['args', 'b', 'c']))
    self.assertEqual(f(0, b=1, c=0).default_args, set(['args']))
    self.assertEqual(f(0, b=1, c=0, d=2).default_args, set(['args']))

    x = f(0)
    self.assertEqual(x.default_args, set(['args', 'b', 'c']))
    x.args = [1]
    self.assertEqual(x.default_args, set(['b', 'c']))
    x.b = 1
    self.assertEqual(x.default_args, set(['c']))
    x.rebind(c=0)
    self.assertEqual(x.default_args, set())

  def test_bound_args(self):
    @pg_functor
    def f(a, *args, b=0, c=1, **kwargs):
      del a, args, b, c, kwargs

    self.assertEqual(f.partial().bound_args, set(['args', 'b', 'c']))
    self.assertEqual(f(1).bound_args, set(['a', 'args', 'b', 'c']))
    self.assertEqual(f(1, 2).bound_args, set(['a', 'args', 'b', 'c']))
    self.assertEqual(f(1, d=2).bound_args, set(['a', 'args', 'b', 'c', 'd']))

    x = f(1, 1, b=1)
    self.assertEqual(x.bound_args, set(['a', 'args', 'b', 'c']))

    x.a = MISSING_VALUE
    self.assertEqual(x.bound_args, set(['args', 'b', 'c']))

    # Setting args to MISSING_VALUE will make it [], thus still considered
    # bound.
    x.args = MISSING_VALUE
    self.assertEqual(x.bound_args, set(['args', 'b', 'c']))

    # Restore x.b to its default value.
    x.rebind(b=MISSING_VALUE)
    self.assertEqual(x.b, 0)
    self.assertEqual(x.bound_args, set(['args', 'b', 'c']))

  def test_unbound_args(self):
    @pg_functor
    def f(a, *args, b=0, c=1, **kwargs):
      del a, args, b, c, kwargs

    self.assertEqual(f.partial().unbound_args, set(['a']))
    self.assertEqual(f(1).unbound_args, set())
    self.assertEqual(f(1, 2).unbound_args, set())
    self.assertEqual(f(1, d=2).unbound_args, set())

    x = f(1, 1, b=1)
    self.assertEqual(x.unbound_args, set())

    x.a = MISSING_VALUE
    self.assertEqual(x.unbound_args, set(['a']))

    # Restore x.b to its default value.
    x.rebind(b=MISSING_VALUE)
    self.assertEqual(x.b, 0)
    self.assertEqual(x.unbound_args, set(['a']))

  def test_is_fully_bound(self):
    @pg_functor
    def f(a, *args, b=0, c=1, **kwargs):
      del a, args, b, c, kwargs

    self.assertFalse(f.partial().is_fully_bound)
    self.assertTrue(f(1).is_fully_bound)
    self.assertTrue(f(1, 0).is_fully_bound)
    self.assertTrue(f(1, 0, b=MISSING_VALUE).is_fully_bound)
    self.assertTrue(f(1, 0, c=MISSING_VALUE).is_fully_bound)
    self.assertTrue(f(1, 0, d=MISSING_VALUE).is_fully_bound)

    x = f(1, 2, 3, b=4, c=5, d=6)

    x.d = MISSING_VALUE
    self.assertTrue(x.is_fully_bound)

    del x.c
    self.assertTrue(x.is_fully_bound)

    x.rebind(args=MISSING_VALUE)
    self.assertEqual(x.args, [])
    self.assertTrue(x.is_fully_bound)

    x.a = MISSING_VALUE
    self.assertFalse(x.is_fully_bound)

  def test_custom_base_class(self):

    class MyFunctor(Functor):
      pass

    @pg_functor(base_class=MyFunctor)
    def my_fun():
      return 0

    self.assertIsInstance(my_fun(), MyFunctor)

  def test_non_default_values(self):
    @pg_functor([
        ('c', pg_typing.Dict([
            ('p', pg_typing.Int(default=1)),
            ('q', pg_typing.Bool(default=True))
        ]))
    ])
    def f(a, b=1, c=Dict(p=1, q=True)):
      del a, b, c

    x = f(1, 1, dict(p=False))
    self.assertEqual(x.non_default_values(), {'a': 1, 'c.p': False})
    self.assertEqual(
        x.non_default_values(flatten=False), {'a': 1, 'c': {'p': False}})

  def test_missing_values(self):
    @pg_functor
    def f(x, y, z=1):
      del x, y, z

    @pg_members([
        ('x', pg_typing.Int())
    ])
    class A(Object):
      pass

    x = f.partial(1)
    # Unbound argument `y` is not considered missing.
    self.assertEqual(x.missing_values(), {})

    # Bound argument `x` with partial object is considered missing.
    x = f.partial(x=A.partial())
    self.assertEqual(x.missing_values(), {'x.x': MISSING_VALUE})

  def test_auto_call(self):
    @pg_functor
    def f(x, y):
      return x + y

    with flags.auto_call_functors():
      self.assertEqual(f(1, 2), 3)
      with self.assertRaisesRegex(
          TypeError, '.* missing 1 required positional argument'):
        _ = f(1)  # pylint: disable=no-value-for-parameter

    self.assertIsInstance(f(1, 2), Functor)


if __name__ == '__main__':
  unittest.main()
